# (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307
# USA.

__revision__ = '$Id: distributors.py,v 1.13 2002/10/16 12:29:43 nico Exp $'

from interfaces import DistributorInterface
import math

def make_new_domains(domains):
    d = {}
    for k,v in domains.items():
        d[k] = v.copy()
    return d

class AbstractDistributor(DistributorInterface):

    def __init__(self, nb_subspaces=2):
        self.nb_subspaces = nb_subspaces
        self.verbose = 0
        
    def findSmallestDomain(self, domains):
        """returns the variable having the smallest domain.
        (or one of such varibles if there is a tie)
        """
        list = [(d.size(),s) for s,d in domains.items() if d.size()>1]
        list.sort()
        return list[0][1]

    def findLargestDomain(self, domains):
        """returns the variable having the largest domain.
        (or one of such variables if there is a tie)
        """
        list = [(d.size(),s) for s,d in domains.items() if d.size()>1]
        list.sort()
        return list[-1][1]

    def nb_subdomains(self, domains):
        return self.nb_subspaces

    def distribute(self,domains,verbose):
        """do the minimal job and let concrete class distribute variables
        """
        self.verbose = verbose
        replicas = []
        for i in range(self.nb_subdomains(domains)):
            replicas.append(make_new_domains(domains))
        modified_domains = self._distribute(*replicas)
        for domain in modified_domains:
            domain.resetFlags()
        return replicas

    def _distribute(self, *args):
        """ method to implement in concrete class

        take self.nb_subspaces copy of the original domains as argument
        distribute the domains and return each modified domain
        """
        raise NotImplementedError()
        
class NaiveDistributor(AbstractDistributor):
    """distributes domains by splitting the smallest domain in 2 new domains
    The first new domain has a size of one,
    and the second has all the other values"""

    def __init__(self):
        AbstractDistributor.__init__(self)
        
    def _distribute(self, d1, d2):
        v = self.findSmallestDomain(d1)
        values = d1[v].getValues()
        if self.verbose:
            print 'Distributing domain for variable',v,'at value',values[0]
        for val in values[1:]:
            d1[v].removeValue(val)
        d2[v].removeValue(values[0])
        return (d1[v],d2[v])

class SplitDistributor(AbstractDistributor):
    """distributes domains by splitting the smallest domain in
    nb_subspaces equal parts or as equal as possible.
    If nb_subspaces is 0, then the smallest domain is split in
    domains of size 1"""
    
    def __init__(self,nb_subspaces=3):
        AbstractDistributor.__init__(self,nb_subspaces)

    def nb_subdomains(self,domains):
        self.__to_split = self.findSmallestDomain(domains)
        if self.nb_subspaces:
            return min(self.nb_subspaces,len(domains[self.__to_split].getValues()))
        else:
            return len(domains[self.__to_split].getValues())
    
    def _distribute(self, *args):
        v = self.__to_split
        nb_subspaces = len(args)
        values = args[0][v].getValues()
        nb_elts = max(1,len(values)*1./nb_subspaces)
        slices = [(int(math.floor(a*nb_elts)),int(math.floor((a+1)*nb_elts)))
                  for a in range(nb_subspaces)]
        if self.verbose:
            print 'Distributing domain for variable',v
        slice = 0
        modified = []
        for dom in args:
            for val in values[:slices[slice][0]]:
                dom[v].removeValue(val) 
            for val in values[slices[slice][1]:]:

                dom[v].removeValue(val)
            modified.append(dom[v])
            slice += 1
        return modified

class DichotomyDistributor(SplitDistributor):
    """distributes domains by splitting the smallest domain in
    two equal parts or as equal as possible"""
    def __init__(self):
        SplitDistributor.__init__(self,2)


class EnumeratorDistributor(SplitDistributor):
    """distributes domains by splitting the smallest domain
    in domains of size 1."""
    def __init__(self):
        SplitDistributor.__init__(self,0)

DefaultDistributor = DichotomyDistributor
