# (c) 2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307
# USA.
"""Tools to work with finite domain variables and constraints

This module provides the following usable classes:
 * FiniteDomain: a class for storing FiniteDomains
 * Expression: a constraint represented as an expression
 * BinaryExpression: a binary constraint represented as an expression
 * various BasicConstraint classes

The Expression and BinaryExpression classes can be constructed using the
make_expression factory function.  """

__revision__ = '$Id: fd.py,v 1.14 2002/10/16 15:45:37 syt Exp $'

from propagation import AbstractDomain,BasicConstraint, ConsistencyFailure
from interfaces import ConstraintInterface
import operator
from psyco_wrapper import psyobj

class FiniteDomain(AbstractDomain):
    def __init__(self,values):
        """values is a list of values in the domain
        This class uses a dictionnary to make sure that there are
        no duplicate values"""
        assert len(values) > 0
        AbstractDomain.__init__(self)
        self._values = {}
        for v in values:
            self._values[v] = 0
        #self.getValues = self._values.keys
        
    def removeValue(self,value):
        del self._values[value]
        self._valueRemoved()
    def size(self):
        return len(self._values)
    def getValues(self):
        return self._values.keys()
    def copy(self):
        return FiniteDomain(self.getValues())
    def __repr__(self):
        return '<FiniteDomain %s>'%str(self.getValues())

##
## Constraints
##    
class Expression(ConstraintInterface,psyobj):
    """A constraint represented as a python expression."""

    def __init__(self,variables,formula,type='fd.Expression'):
        """variables is a list of variables which appear in the formula
        formula is a python expression that will be evaluated as a boolean"""

        self._variables = variables
        self.formula = formula
        self.type = type
        self.filterFunc = eval('lambda %s: %s'%(','.join(variables),formula),{},{})

    def affectedVariables(self):
        """ Return a list of all variables affected by this constraint """
        return self._variables

    def estimateCost(self,domains):
        """ Return an estimate of the cost of the narrowing of the constraint """
        return reduce(operator.mul,[domains[v].size() for v in self.affectedVariables()])

    def isVariableRelevant(self,variable):
        return variable in self._variables

    def _init_result_cache(self,variables):
        """key = (variable,value), value = [has_success,has_failure]"""
        result_cache = {}
        for var_name,values, offset in variables:
            for val in values:
                result_cache[(var_name,val)] = [0,0]
            
        return result_cache

    def narrow(self,domains):
        maybe_entailed = 1
        ffunc = self.filterFunc
        
        # sort variables to instanciate those with fewer possible values first
        variables = [(len(domains[k].getValues()),
                      [k,domains[k].getValues(),0])
                     for k in self._variables]
        
        variables.sort()
        variables = [v[1] for v in variables]
        go_on = 1
        kwargs = {}
        result_cache = self._init_result_cache(variables)
        # initialize the keyword arguments for the first call
        for var_name,values,offset in variables:
            kwargs[var_name]=values[offset]
        # iterate for all values
        while go_on:
            if apply(ffunc,(),kwargs):
                for var,val in kwargs.items():
                    result_cache[(var,val)][0]=1
            else:
                for var,val in kwargs.items():
                    result_cache[(var,val)][1]=1
            # try to instanciate the next variable
            for curr in variables:
                if curr[2]+1<len(curr[1]):
                    curr[2] +=1
                    kwargs[curr[0]] = curr[1][curr[2]]
                    break
                else:
                    curr[2] = 0
                    kwargs[curr[0]] = curr[1][0]
            else:
                # it's over
                go_on = 0
        try:
            while 1:
                (var,val),(has_success,has_failure) = result_cache.popitem()
                if not has_success:
                    domains[var].removeValue(val)
                if has_failure and has_success:
                    maybe_entailed = 0
        except ConsistencyFailure, e:
            raise ConsistencyFailure('Inconsistency while applying %s'%repr(self))
        except KeyError:
            # There are no more value in result_cache
            pass

        return maybe_entailed

    def __repr__(self):
        return '<%s "%s">'%(self.type, self.formula)

class BinaryExpression(Expression):
    """A binary constraint represented as a python expression

    This implementation uses a narrowing algorithm optimized for
    binary constraints."""
    
    def __init__(self,variables,formula,type='fd.BinaryExpression'):
        assert len(variables) == 2
        Expression.__init__(self,variables,formula,type)

    def narrow(self,domains):
        maybe_entailed = 1
        var1 = self._variables[0]
        dom1 = domains[var1]
        values1 = dom1.getValues()
        var2 = self._variables[1]
        dom2 = domains[var2]
        values2 = dom2.getValues()
        ffunc = self.filterFunc
        if dom2.size() < dom1.size():
            var1,var2 = var2,var1
            dom1,dom2 = dom2,dom1
            values1,values2 = values2, values1
            
        kwargs = {}
        result1 = {}
        for v in values1:
            result1[v] = [0,0]
        result2 = {}
        for v in values2:
            result2[v] = [0,0]

        try:
            # iterate for all values
            for val1 in dom1.getValues():
                kwargs[var1] = val1
                for val2 in dom2.getValues():
                    kwargs[var2] = val2
                    if apply(ffunc,(), kwargs):
                        result1[val1][0]=1
                        result2[val2][0]=1
                    else:
                        result1[val1][1]=1
                        result2[val2][1]=1
            for val1,(has_success,has_failure) in result1.items():
                if not has_success:
                    dom1.removeValue(val1)
                if has_failure :#and has_success:
                    maybe_entailed = 0
            for val2,(has_success,has_failure) in result2.items():
                if not has_success:
                    dom2.removeValue(val2)
                if has_failure :#and has_success:
                    maybe_entailed = 0
        except ConsistencyFailure:
            raise ConsistencyFailure('Inconsistency while applying %s'%repr(self))
        except Exception,e:
            print self, kwargs
            raise e
        return maybe_entailed


def make_expression(variables,formula,constraint_type=None):
    # encode unicode
    vars = []
    for var in variables:
        if type(var) == type(u''):
            vars.append(var.encode())
        else:
            vars.append(var)
    if len(vars) == 2:
        if constraint_type is not None:
            return BinaryExpression(vars,formula,constraint_type)
        else:
            return BinaryExpression(vars,formula)

    else:
        if constraint_type is not None:
            return Expression(vars,formula,constraint_type)
        else:
            return Expression(vars,formula)


class Equals(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_equals)

class NotEquals(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_neq)

class LesserThan(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_lt)

class LesserOrEqual(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_lte)

class GreaterThan(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_gt)

class GreaterOrEqual(BasicConstraint):
    def __init__(self,variable,reference):
        BasicConstraint.__init__(self,variable,reference,_gte)

def _equals(x,y):
    return x == y

def _lt(x,y):
    return x < y

def _lte(x,y):
    return x <= y

def _gt(x,y):
    return x > y

def _gte(x,y):
    return x >= y

def _neq(x,y):
    return x != y




