# (c) 2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307
# USA.

__revision__ = '$Id: propagation.py,v 1.21 2002/10/16 15:45:37 syt Exp $'

from interfaces import *
from __future__ import generators
import bisect
from time import strftime
from psyco_wrapper import psyobj

class ConsistencyFailure(Exception):
    pass

class Repository(psyobj):
    """Stores variables, domains and constraints
    Propagates domain changes to constraints
    Manages the constraint evaluation queue"""
    def __init__(self,variables,domains,
                 constraints = None):
        # encode unicode
        for i in range(len(variables)):
            if type(variables[i]) == type(u''):
                variables[i] = variables[i].encode()
                
        self._variables = variables   # list of variable names
        self._domains = domains    # maps variable name to domain object
        self._constraints = [] # list of constraint objects
        self._queue = []       # queue of constraints waiting to be processed
        self._variableListeners = {}
        for v in self._variables:
            self._variableListeners[v] = []
            assert self._domains.has_key(v)
        for c in constraints or ():
            self.addConstraint(c)

    def __repr__(self):
        return '<Repository nb_constraints=%d domains=%s>'%(len(self._constraints),self._domains)

    def vcg_draw(self, file, title='Constraints graph'):
        """ draw a constraints graph readable by vcg
        """
        from logilab.pyreverse.extensions.vcgutils import VCGPrinter, EDGE_ATTRS
        f = open(file, 'w+')
        self.printer = VCGPrinter(f)        
        self.printer.open_graph(title=title, textcolor='black'
#                                layoutalgorithm='dfs',
#                               manhattan_edges='yes'
#                               port_sharing='no'
#                                late_edge_labels='yes'
                                )
        # print variables (nodes)
        for var in self._variables:
            self.printer.node(var, shape='ellipse')
        # print constraints (edges)
        type_colors = {}
        color_index = 2
        i = 0
        for constraint in self._constraints:
            key = constraint.type
            if not type_colors.has_key(key):
                type_colors[key] = color_index
                color_index += 1
            affected_vars = constraint.affectedVariables()
            if len(affected_vars) <= 1:
                continue
            if len(affected_vars) == 2:
                var1 = affected_vars[0]
                var2 = affected_vars[1]
                self.printer.edge(var1, var2, arrowstyle='none',
                                  color=EDGE_ATTRS['color'][type_colors[key]])
                continue
            n_id = 'N_aire%i'%i
            i+=1
            n = self.printer.node(n_id, shape='triangle')
            for var1 in affected_vars:
                self.printer.edge(var1, n_id, arrowstyle='none',
                                  color=EDGE_ATTRS['color'][type_colors[key]])
        # print legend
        for type, color in type_colors.items():
            self.printer.node(type, shape='box',
                              color=EDGE_ATTRS['color'][color])
        self.printer.close_graph()
        f.close()
        
    def addConstraint(self,constraint):
        if isinstance(constraint,BasicConstraint):
            # Basic constraints are processed just once
            # because they are straight away entailed
            v = constraint.getVariable()
            constraint.narrow({v:self._domains[v]})
        else:
            self._constraints.append(constraint)
            for v in constraint.affectedVariables():
                self._variableListeners[v].append(constraint)
            bisect.insort(self._queue,
                          (constraint.estimateCost(self._domains),constraint))

    def _removeConstraint(self,constraint):
        self._constraints.remove(constraint)
        for v in constraint.affectedVariables():
            self._variableListeners[v].remove(constraint)
        for i in range(len(self._queue)):
            if self._queue[i][1] is constraint:
                del self._queue[i]
                break

    def getDomains(self):
        return self._domains

    def distribute(self,distributor,verbose=0):
        return [Repository(self._variables,d,self._constraints) 
                for d in distributor.distribute(self._domains,verbose)]
    
    def consistency(self,verbose=0):
        """Prunes the domains of the variables
        This method calls constraint.narrow() and queues constraints
        that are affected by recent changes in the domains"""
        if verbose:
            print strftime('%H:%M:%S'),'** Consistency **'
        while self._queue:
            if verbose>2:
                print strftime('%H:%M:%S'),'Queue',self._queue
            cost,constraint = self._queue.pop(0)
            if verbose > 1:
                print strftime('%H:%M:%S'), 'Trying to entail constraint',constraint,'[cost:%d]'%cost
            entailed = constraint.narrow(self._domains)
            for v in constraint.affectedVariables():
                d = self._domains[v]
                if d.hasChanged():
                    if verbose>1 :
                        print strftime('%H:%M:%S'),' -> New domain for variable',v,'is',d
                    for c in self._variableListeners[v]:
                        if c is not constraint:
                            for c_cost,cst in self._queue:
                                if cst is c:
                                    break
                            else:
                                bisect.insort(self._queue,
                                              (c.estimateCost(self._domains),c))
                    d.resetFlags()
            if entailed:
                if verbose:
                    print strftime('%H:%M:%S'),"--> Entailed constraint",constraint
                self._removeConstraint(constraint)

class Solver(psyobj):
    def __init__(self,distributor=None):
        if distributor is None:
            from distributors import DefaultDistributor
            distributor = DefaultDistributor()
        self._distributor = distributor
        self.verbose = 1
        self.max_depth = 0

    def solve_one(self, repository, verbose=0):
        """Generates only one solution"""
        self.verbose = verbose
        self.max_depth = 0
        try:
            return self._solve(repository).next()
        except StopIteration, e:
            return
        
    def solve_best(self, repository, cost_func, verbose=0):
        """Generates solution with an improving cost"""
        self.verbose = verbose
        self.max_depth = 0
        best_cost = None
        for s in self._solve(repository):
            cost = cost_func(s)
            if best_cost is None or cost <= best_cost:
                best_cost = cost
                yield s,cost            
        
    def solve_all(self, repository, verbose=0):
        """Generates all solutions"""
        self.verbose = verbose
        self.max_depth = 0
        for s in self._solve(repository):
            yield s

    def solve(self,repository,verbose=0):
        self.max_depth = 0
        solutions = []
        for s in self.solve_all(repository,verbose):
            solutions.append(s)
        return solutions
        
    def _solve(self,repository,recursion_level=0 ):
        """main generator"""
        if recursion_level > self.max_depth:
            self.max_depth = recursion_level
        if self.verbose:
            print strftime('%H:%M:%S'),'*** [%d] Solve called with repository'%recursion_level,repository
        try:
            repository.consistency(self.verbose)
        except ConsistencyFailure,e:
            if self.verbose:
                print strftime('%H:%M:%S'),e
            pass
        else:
            if self._foundSolution(repository.getDomains().values()):
                s = {}
                for v,d in repository.getDomains().items():
                    s[v] = d.getValues()[0]
                if self.verbose:
                    print strftime('%H:%M:%S'),'### Found Solution',s
                    print '-'*80
                yield s                    
            else:
                for r in repository.distribute(self._distributor,self.verbose):
                    for s in self._solve(r,recursion_level+1):
                        if s is not None:
                            yield s
        if recursion_level == 0 and self.verbose:
            print strftime('%H:%M:%S'),'Finished search'
            print strftime('%H:%M:%S'),'Maximum recursion depth = ',self.max_depth


        
    def _foundSolution(self,domains):
        for d in domains:
            if d.size()!=1:
                return 0
        return 1
        




class BasicConstraint(ConstraintInterface):
    """A BasicConstraint, which is never queued by the Repository
    A BasicConstraint affects only one variable, and will be entailed
    on the first call to narrow()"""
    def __init__(self,variable,reference,operator):
        """variables is a list of variables on which
        the constraint is applied"""
        self._variable = variable
        self._reference = reference
        self._operator = operator

    def __repr__(self):
        return '<%s %s %s>'% (self.__class__, self._variable, self._reference)

    def isVariableRelevant(self,variable):
        return variable == self._variable

    def estimateCost(self,domains):
        return 0 # get in the first place in the queue
    
    def affectedVariables(self):
        return [self._variable]
    
    def getVariable(self):
        return self._variable
        
    def narrow(self,domains):
        """ensures that the domains are consistent with the constraint
        Calls domain.removeValue to remove values from a domain
        raises FailedConstraint if the narrowing fails"""
        domain = domains[self._variable]
        op = self._operator
        ref = self._reference
        try:
            for v in domain.getValues():
                if not op(v,ref):
                    domain.removeValue(v)
        except ConsistencyFailure, e:
            raise ConsistencyFailure('inconsistency while applying %s'%repr(self))
        return 1


class AbstractDomain(DomainInterface):
    """Implements the functionnality related to the changed flag.
    Can be used as a starting point for concrete domains"""
    def __init__(self):
        self.__changed = 0
    def resetFlags(self):
        self.__changed = 0
    
    def hasChanged(self):
        return self.__changed

    def _valueRemoved(self):
        """The implementation of removeValue should call this method"""
        self.__changed = 1
        if self.size() == 0:
            raise ConsistencyFailure()
    
